from io import BytesIO
from itertools import chain
from typing import Union

from attr import attrs

from av_clipboard_lib.base_types import PositionValue, RowPosition, STRUCT_BYTE, STRUCT_CHAR, STRUCT_DOUBLE, \
    STRUCT_DWORD, TimePosition
from av_clipboard_lib.varint import decode_next_varint, encode_varint


def _register_to(registry):
    def register(kind):
        def _(cls):
            registry[kind] = cls
            registry[cls] = kind
            return cls

        return _

    return register


NOTE_REGISTRY = {}
STRUCTURE_REGISTRY = {}

_register_note = _register_to(NOTE_REGISTRY)
_register_structure = _register_to(STRUCTURE_REGISTRY)


@attrs(auto_attribs=True)
class InstantNote:
    column: int
    position: PositionValue

    @classmethod
    def from_triplet(cls, column, first, _):
        return cls(column, first)

    @property
    def order_tuple(self):
        """Return a pair that can be used to sort data"""
        return self.position, self.column

    @property
    def encoded(self):
        return b''.join((
            STRUCT_BYTE.pack(self.column | 0x80),
            self.position.encoded,
            self.position.encoded,
            STRUCT_BYTE.pack(NOTE_REGISTRY[self.__class__]),
        ))


@attrs(auto_attribs=True)
class LongNote:
    column: int
    start_position: PositionValue
    end_position: PositionValue

    @classmethod
    def from_triplet(cls, column, first, second):
        return cls(column, first, second)

    @property
    def order_tuple(self):
        """Return a pair that can be used to sort data"""
        return self.start_position, self.column

    @property
    def encoded(self):
        return b''.join((
            STRUCT_BYTE.pack(self.column | 0x80),
            self.start_position.encoded,
            self.end_position.encoded,
            STRUCT_BYTE.pack(NOTE_REGISTRY[self.__class__]),
        ))


@_register_note(None)
class Tap(InstantNote):
    @property
    def encoded(self):
        return b''.join((
            STRUCT_BYTE.pack(self.column),
            self.position.encoded,
        ))


@_register_note(0x00)
class Hold(LongNote): pass


@_register_note(0x01)
class Mine(InstantNote): pass


@_register_note(0x02)
class Roll(LongNote): pass


@_register_note(0x03)
class Lift(InstantNote): pass


@_register_note(0x04)
class Fake(InstantNote): pass


def generate_structure_serializer_pair(order, packers):
    @classmethod
    def decode(cls, stream: BytesIO):
        row = RowPosition.decode_next_as_dword(stream)
        data = {
            name: packer.unpack(stream.read(packer.size))[0]
            for name, packer in zip(order, packers)
        }

        return cls(row, **data)

    @property
    def encoded(self):
        return b''.join((
            self.position.encoded_as_dword,
            *(
                packer.pack(getattr(self, name))
                for name, packer in zip(order, packers)
            )
        ))

    return decode, encoded


@attrs(auto_attribs=True)
@_register_structure(0x00)
class BPM:
    position: RowPosition
    bpm: float

    decode, encoded = generate_structure_serializer_pair(['bpm'], [STRUCT_DOUBLE])


@attrs(auto_attribs=True)
@_register_structure(0x01)
class Stop:
    position: RowPosition
    time: float

    decode, encoded = generate_structure_serializer_pair(['time'], [STRUCT_DOUBLE])


@attrs(auto_attribs=True)
@_register_structure(0x02)
class Delay:
    position: RowPosition
    time: float

    decode, encoded = generate_structure_serializer_pair(['time'], [STRUCT_DOUBLE])


@attrs(auto_attribs=True)
@_register_structure(0x03)
class Warp:
    position: RowPosition
    skipped_rows: int

    decode, encoded = generate_structure_serializer_pair(['skipped_rows'], [STRUCT_DWORD])


@attrs(auto_attribs=True)
@_register_structure(0x04)
class TimeSignature:
    position: RowPosition
    numerator: int
    denominator: int

    decode, encoded = generate_structure_serializer_pair(['numerator', 'denominator'], [STRUCT_DWORD] * 2)


@attrs(auto_attribs=True)
@_register_structure(0x05)
class Ticks:
    position: RowPosition
    ticks: int

    decode, encoded = generate_structure_serializer_pair(['ticks'], [STRUCT_DWORD])


@attrs(auto_attribs=True)
@_register_structure(0x06)
class Combo:
    position: RowPosition
    combo_mul: int
    miss_mul: int

    decode, encoded = generate_structure_serializer_pair(['combo_mul', 'miss_mul'], [STRUCT_DWORD] * 2)


@attrs(auto_attribs=True)
@_register_structure(0x07)
class Speed:
    position: RowPosition
    ratio: float
    delay: float
    delay_is_time: bool

    @classmethod
    def decode(cls, stream: BytesIO):
        row = RowPosition.decode_next_as_dword(stream)
        ratio, = STRUCT_DOUBLE.unpack(stream.read(8))
        delay, = STRUCT_DOUBLE.unpack(stream.read(8))
        is_time, = STRUCT_DWORD.unpack(stream.read(4))

        return cls(row, ratio, delay, bool(is_time))

    @property
    def encoded(self):
        return b''.join((
            self.position.encoded_as_dword,
            STRUCT_DOUBLE.pack(self.ratio),
            STRUCT_DOUBLE.pack(self.delay),
            STRUCT_DWORD.pack(int(self.delay_is_time))
        ))


@attrs(auto_attribs=True)
@_register_structure(0x08)
class Scroll:
    position: RowPosition
    ratio: float

    decode, encoded = generate_structure_serializer_pair(['ratio'], [STRUCT_DOUBLE])


@attrs(auto_attribs=True)
@_register_structure(0x09)
class FakeSegment:
    position: RowPosition
    fake_rows_amt: int

    decode, encoded = generate_structure_serializer_pair(['fake_rows_amt'], [STRUCT_DWORD])


@attrs(auto_attribs=True)
@_register_structure(0x0A)
class Label:
    position: RowPosition
    message: str

    @classmethod
    def decode(cls, stream: BytesIO):
        row = RowPosition.decode_next_as_dword(stream)
        message_len = decode_next_varint(stream)
        unpacked_message = STRUCT_CHAR.iter_unpack(stream.read(message_len))
        message = b''.join(chain.from_iterable(unpacked_message)).decode('ascii')

        return cls(row, message)

    @property
    def encoded(self):
        return b''.join((
            self.position.encoded_as_dword,
            encode_varint(len(self.message)),
            self.message.encode('ascii')
        ))


NoteType = Union[Tap, Hold, Mine, Roll, Lift, Fake]
StructureType = Union[BPM, Stop, Delay, Warp, TimeSignature, Ticks, Combo, Speed, Scroll, FakeSegment, Label]


def decode_next_note(stream: BytesIO, is_time: bool) -> NoteType:
    first_byte, = STRUCT_BYTE.unpack(stream.read(1))
    first_position = is_time and TimePosition.decode_next(stream) or RowPosition.decode_next_as_varint(stream)

    if not first_byte & 0x80:
        return NOTE_REGISTRY[None](first_byte, first_position)

    column = first_byte ^ 0x80
    second_position = is_time and TimePosition.decode_next(stream) or RowPosition.decode_next_as_varint(stream)
    kind, = STRUCT_BYTE.unpack(stream.read(1))

    return NOTE_REGISTRY[kind].from_triplet(column, first_position, second_position)


def decode_next_structure(stream: BytesIO, kind: int) -> StructureType:
    return STRUCTURE_REGISTRY[kind].decode(stream)



from av_clipboard_lib import clipboard_data
from av_clipboard_lib import av_objects
from av_clipboard_lib import base85
from av_clipboard_lib import varint

from av_clipboard_lib.clipboard_data import parse_av_clipboard_data, produce_av_clipboard_data
from av_clipboard_lib.av_objects import (
    Tap, Hold, Mine, Roll, Lift, Fake,
    BPM, Stop, Delay, Warp, TimeSignature, Ticks, Combo,
    Speed, Scroll, FakeSegment, Label
)
from av_clipboard_lib.base_types import RowPosition, TimePosition



from io import BytesIO

from av_clipboard_lib.av_objects import BPM, Combo, Delay, FakeSegment, Hold, Label, Mine, Roll, Scroll, Speed, Tap, \
    Ticks, \
    TimeSignature, Warp, \
    decode_next_note
from av_clipboard_lib.base_types import RowPosition, TimePosition
from av_clipboard_lib.clipboard_data import RowCopy, StructureCopy, parse_av_clipboard_data, produce_av_clipboard_data

P = RowPosition(58301)
P_hex = 'BDC703'
PW_hex = 'BDE30000'
DW = 59311
DW_hex = 'AFE70000'
D = TimePosition(3.14)
DD = 3.14
D_hex = '1F85EB51B81E0940'


def make_stream(hex_bytes):
    return BytesIO(bytes.fromhex(hex_bytes))


class TestAVObject:
    @staticmethod
    def generic_test(target, source_func, hex_bytes, *args):
        assert source_func(make_stream(hex_bytes), *args) == target
        assert target.encoded == bytes.fromhex(hex_bytes)

    def test_note_tap(self):
        self.generic_test(
            Tap(column=2, position=P),
            decode_next_note,
            f'02 {P_hex}',
            False
        )

    def test_note_roll(self):
        self.generic_test(
            Roll(column=1, start_position=P, end_position=P),
            decode_next_note,
            f'81 {P_hex} {P_hex} 02',
            False
        )

    def test_timed_tap(self):
        self.generic_test(
            Tap(column=2, position=D),
            decode_next_note,
            f'02 {D_hex}',
            True
        )

    def test_structure_double(self):
        self.generic_test(
            BPM(position=P, bpm=DD),
            BPM.decode,
            f'{PW_hex} {D_hex}',
        )

    def test_structure_dword(self):
        self.generic_test(
            Warp(position=P, skipped_rows=DW),
            Warp.decode,
            f'{PW_hex} {DW_hex}',
        )
        assert Warp.decode(make_stream(f'{PW_hex} {DW_hex}')) == Warp(position=P, skipped_rows=DW)

    def test_structure_speed(self):
        self.generic_test(
            Speed(position=P, ratio=DD, delay=DD, delay_is_time=True),
            Speed.decode,
            f'{PW_hex} {D_hex} {D_hex} 01000000',
        )

    def test_structure_label(self):
        self.generic_test(
            Label(position=P, message="gamma"),
            Label.decode,
            f'{PW_hex} 05 {"gamma".encode("ascii").hex()}',
        )


class TestLib:
    def test_av_string_structure(self):
        input_string = (
            "4172726f77566f727465783a74656d706f3"
            "a21575733237a7a3a2d5d33667a7a44456e34282"
            "c514966452a456c75386862573e75213c592255"
            "212121452d21212124263b75636d75226f6e572"
            "7236c6a722a213c6e2c56212123255c21212124"
            "285e5d343f372e4b42474b555d43477121584a5"
            "7272121222a706c56592f35292a6c486a62666e"
            "3b5426337033712121222c422121222a706c565"
            "92f3523736532343e605a62702f336a34392121"
            "21242a3a5d554f7234555466394534"
            "6d5f64213d3d38572121254f6a212121242c59513"
            "45f282a4362305d4345524a2a2b435c6271436961237"
            "04345526837403c3c572b46212c523c426c6137"
        )

        av = bytes.fromhex(input_string).decode('ascii')
        target = StructureCopy(
            objects=[
                BPM(position=RowPosition(row=0), bpm=60.0), BPM(position=RowPosition(row=12), bpm=240.0),
                Delay(position=RowPosition(row=36), time=0.666),
                Warp(position=RowPosition(row=48), skipped_rows=12),
                TimeSignature(position=RowPosition(row=84), numerator=6, denominator=9),
                Ticks(position=RowPosition(row=132), ticks=314),
                Combo(position=RowPosition(row=192), combo_mul=42, miss_mul=420),
                Speed(position=RowPosition(row=252), ratio=6.28, delay=4.2, delay_is_time=False),
                Speed(position=RowPosition(row=288), ratio=3.14, delay=14.48, delay_is_time=True),
                Scroll(position=RowPosition(row=336), ratio=13.37),
                FakeSegment(position=RowPosition(row=384), fake_rows_amt=29568),
                Label(position=RowPosition(row=432), message='Fuck me ballsack what is this')
            ]
        )
        assert produce_av_clipboard_data(parse_av_clipboard_data(av)) == av
        assert parse_av_clipboard_data(av) == target

    def test_av_string_note(self):
        input_string = "4172726f77566f727465783a6e6f7465733a21214539254a4d3862594a6d615a40212f2440365e5d3d4b"
        av = bytes.fromhex(input_string).decode('ascii')

        target = RowCopy(objects=[
            Tap(column=0, position=RowPosition(row=0)),
            Mine(column=1, position=RowPosition(row=48)),
            Hold(column=2, start_position=RowPosition(96), end_position=RowPosition(144)),
            Roll(column=3, start_position=RowPosition(144), end_position=RowPosition(192)),
        ])
        assert produce_av_clipboard_data(parse_av_clipboard_data(av)) == av
        assert parse_av_clipboard_data(av) == target



from io import BytesIO, StringIO

from av_clipboard_lib.base_types import STRUCT_DWORD_U

_KEY = [0x55 ** 4, 0x55 ** 3, 0x55 ** 2, 0x55 ** 1, 0x55 ** 0]


def encode_dwords_to_base85(data: bytes) -> str:
    """Converts `data` into AV clipboard format

    >>> encode_dwords_to_base85(bytes.fromhex("C9E8C919DC2C7E0E"))
    'alphagamma'
    """
    segments = BytesIO()

    padding = (4 - len(data) % 4) % 4
    data = BytesIO(data + b'\x00' * padding)

    for word_data in iter(lambda: data.read(4), b''):
        dword, = STRUCT_DWORD_U.unpack(word_data)

        segments.write(
            bytes(
                0x21 + (dword // key) % 0x55
                for key in _KEY
            )
        )

    string_out = segments.getvalue().decode('ascii')
    if padding:
        string_out = string_out[:-padding]

    return compress_base85(string_out)


def decode_dwords_from_base85(data: str) -> bytes:
    """Convert `data` from AV clipboard format into bytes

    >>> decode_dwords_from_base85('alphagamma').hex().upper()
    'C9E8C919DC2C7E0E'
    """
    byte_array = BytesIO()

    data = decompress_base85(data)

    padding = (5 - len(data) % 5) % 5
    data = StringIO(data + chr(0x21 + 0x55) * padding)

    for segment in iter(lambda: data.read(5), ''):
        dword = sum(
            (ord(datum) - 0x21) * key
            for key, datum in zip(_KEY, segment)
        )
        byte_array.write(STRUCT_DWORD_U.pack(dword))

    bytes_out = byte_array.getvalue()
    if padding:
        bytes_out = bytes_out[:-padding]

    return bytes(bytes_out)


def decompress_base85(data: str) -> str:
    """Replace "z" with equivalent "!!!!!" in `data`."""

    return data.replace('z', '!!!!!')


def compress_base85(data: str) -> str:
    """Replace null dwords to "z" character"""

    data = StringIO(data)

    return ''.join(
        datum == '!!!!!' and 'z' or datum
        for datum in iter(lambda: data.read(5), '')
    )



from io import BytesIO
from struct import Struct
from typing import Union

from attr import attrs

from av_clipboard_lib.varint import decode_next_varint, encode_varint

STRUCT_BYTE = Struct('<B')
STRUCT_CHAR = Struct('<c')
STRUCT_DOUBLE = Struct('<d')
STRUCT_DWORD = Struct('<I')
STRUCT_DWORD_U = Struct('>I')


@attrs(auto_attribs=True, eq=True, order=True)
class RowPosition:
    row: int

    @property
    def encoded_as_varint(self):
        return encode_varint(self.row)

    encoded = encoded_as_varint

    @property
    def encoded_as_dword(self):
        return STRUCT_DWORD.pack(self.row)

    @classmethod
    def decode_next_as_varint(cls, stream: BytesIO):
        return cls(decode_next_varint(stream))

    @classmethod
    def decode_next_as_dword(cls, stream: BytesIO):
        return cls(*STRUCT_DWORD.unpack(stream.read(4)))


@attrs(auto_attribs=True, eq=True, order=True)
class TimePosition:
    seconds: float

    @property
    def encoded(self):
        return STRUCT_DOUBLE.pack(self.seconds)

    @classmethod
    def decode_next(cls, stream: BytesIO):
        return cls(*STRUCT_DOUBLE.unpack(stream.read(8)))


PositionValue = Union[RowPosition, TimePosition]



from io import BytesIO
from itertools import count


def decode_next_varint(stream: BytesIO) -> int:
    """Read the var int in `stream`, mutating it. Inverse of `encode_varint`

    >>> decode_next_varint(BytesIO(bytes.fromhex("BDC703F00DBAADF00DBAAD")))
    58301
    """

    result = 0

    for segment_num in count():
        next_byte = int.from_bytes(stream.read(1), 'big')
        segment = next_byte & 0x7F
        result += segment << (7 * segment_num)

        if not next_byte & 0x80:
            break

    return result


def encode_varint(number: int) -> bytes:
    """Converts `number` into a varint bytes representation. Inverse of `decode_next_varint`.

    >>> encode_varint(58301).hex().upper()
    'BDC703'
    """
    if number < 0:
        raise ValueError('Argument cannot be negative')

    if number == 0:
        return b'\x00'

    byte_segments = []
    while number != 0:
        byte_segments.append(number & 0x7F)
        number >>= 7

    *alpha, beta = byte_segments
    return bytes(byte | 0x80 for byte in alpha) + beta.to_bytes(1, 'big')



from setuptools import find_packages, setup

with open("README.md", encoding='utf-8') as in_:
    setup(
        name='av-clipboard-lib',
        version='1.9.1',
        packages=find_packages(),
        url='https://github.com/DeltaEpsilon7787/av-clipboard-lib',
        license='MIT',
        author='DeltaEpsilon7787',
        author_email='deltaepsilon7787@gmail.com',
        long_description=in_.read(),
        long_description_content_type="text/markdown",
        description='Small ArrowVortex clipboard processing library',
        python_requires='>=3.6',
        install_requires=["attrs"],
    )



from io import BytesIO
from itertools import groupby
from operator import attrgetter
from typing import List, Union

from attr import attrs

from av_clipboard_lib.av_objects import NoteType, STRUCTURE_REGISTRY, StructureType, decode_next_note, \
    decode_next_structure
from av_clipboard_lib.base85 import decode_dwords_from_base85, encode_dwords_to_base85
from av_clipboard_lib.base_types import STRUCT_BYTE
from av_clipboard_lib.varint import decode_next_varint, encode_varint


@attrs(auto_attribs=True)
class RowCopy:
    objects: List[NoteType]

    @classmethod
    def decode(cls, data: BytesIO):
        count = decode_next_varint(data)

        return cls([
            decode_next_note(data, False)
            for _ in range(count)
        ])

    @property
    def sorted_objects(self):
        return sorted(self.objects, key=attrgetter('order_tuple'))

    @property
    def encoded(self):
        buffer = BytesIO()

        buffer.write(b'\x00')
        buffer.write(encode_varint(len(self.objects)))
        for datum in self.sorted_objects:
            buffer.write(datum.encoded)

        return buffer.getvalue()


@attrs(auto_attribs=True)
class TimeCopy:
    objects: List[NoteType]

    @classmethod
    def decode(cls, data: BytesIO):
        count = decode_next_varint(data)

        return cls([
            decode_next_note(data, True)
            for _ in range(count)
        ])

    @property
    def sorted_objects(self):
        return sorted(self.objects, key=attrgetter('order_tuple'))

    @property
    def encoded(self):
        buffer = BytesIO()

        buffer.write(b'\x01')
        buffer.write(encode_varint(len(self.objects)))
        for datum in self.sorted_objects:
            buffer.write(datum.encoded)

        return buffer.getvalue()


@attrs(auto_attribs=True)
class StructureCopy:
    objects: List[StructureType]

    @classmethod
    def decode(cls, data: BytesIO):
        objects = []
        count = decode_next_varint(data)
        while count > 0:
            kind, = STRUCT_BYTE.unpack(data.read(1))
            for _ in range(count):
                objects.append(decode_next_structure(data, kind))
            count = decode_next_varint(data)
        return cls(objects)

    @property
    def sorted_objects(self):
        return sorted(self.objects, key=lambda x: (STRUCTURE_REGISTRY[x.__class__], x.position))

    @property
    def encoded(self):
        buffer = BytesIO()

        objects = self.sorted_objects
        object_groups = groupby(objects, key=lambda x: STRUCTURE_REGISTRY[x.__class__])

        for kind, group_objects in object_groups:
            group_objects = [*group_objects]
            buffer.write(encode_varint(len(group_objects)))
            buffer.write(STRUCT_BYTE.pack(kind))
            for datum in group_objects:
                buffer.write(datum.encoded)
        buffer.write(STRUCT_BYTE.pack(0))

        return buffer.getvalue()


CopyType = Union[RowCopy, TimeCopy, StructureCopy]


def parse_av_clipboard_data(data: str) -> CopyType:
    """Transform valid AV clipboard `data` into a specific copy object."""
    is_note_data = data.startswith('ArrowVortex:notes:')
    is_tempo_data = data.startswith('ArrowVortex:tempo:')
    if not (is_note_data or is_tempo_data):
        raise ValueError('Argument is not AV clipboard data')

    data = data[18:]
    data = decode_dwords_from_base85(data)
    data = BytesIO(data)

    if is_note_data:
        is_time_based = bool(*STRUCT_BYTE.unpack(data.read(1)))
        return is_time_based and TimeCopy.decode(data) or RowCopy.decode(data)
    return StructureCopy.decode(data)


def produce_av_clipboard_data(elmns: CopyType) -> str:
    """Converts valid `elmns` into AV clipboard data"""

    typ = type(elmns)
    if typ in {RowCopy, TimeCopy}:
        return f'ArrowVortex:notes:{encode_dwords_to_base85(elmns.encoded)}'
    else:
        return f'ArrowVortex:tempo:{encode_dwords_to_base85(elmns.encoded)}'
